@@ -17,6 +17,7 @@ import importlib
 import json
 import logging
 import time
+import uuid
 from typing import Any
 
 logging.basicConfig(level=logging.INFO)
@@ -74,6 +75,13 @@ class AgentChatRequest(BaseModel):
     timeout: int = 30
 
 
+class BatchTasksRequest(BaseModel):
+    """Request to submit multiple tasks to an agent's queue."""
+    tasks: list[dict[str, Any]]  # List of task payloads
+    wait: bool = False  # Whether to wait for responses
+    batch_id: str | None = None  # Optional: use existing batch_id for trace grouping
+
+
 class AgentInfo(BaseModel):
     """Agent information."""
     name: str
@@ -239,7 +247,7 @@ async def health():
     
     return {
         "status": "ok",
-        "version": "0.8.3",
+        "version": "0.8.4",
         "components": {
             "database": db_type,
             "storage": storage_type,
@@ -624,6 +632,71 @@ async def chat_with_agent(agent_name: str, request: AgentChatRequest):
         raise HTTPException(status_code=500, detail=str(e))
 
 
+@app.post("/api/agents/{agent_name}/batch")
+async def batch_submit_tasks(agent_name: str, request: BatchTasksRequest):
+    """
+    Submit multiple tasks to an agent's queue in parallel.
+    
+    Each task will be distributed to available workers via Redis Consumer Groups.
+    All traces from workers will be grouped under a single batch_id (job_id).
+    
+    Returns immediately with task IDs (non-blocking by default).
+    """
+    try:
+        # Generate batch_id (used as job_id for trace grouping)
+        batch_id = request.batch_id or str(uuid.uuid4())
+        
+        task_ids = []
+        for i, task_payload in enumerate(request.tasks):
+            # Ensure job_id is in payload (critical for trace grouping)
+            task_with_job = {
+                **task_payload,
+                "job_id": batch_id,  # All tasks share the same job_id
+                "_batch_index": i,  # Optional: track task order
+            }
+            
+            task_id = await message_bus.publish_task(agent_name, task_with_job)
+            task_ids.append(task_id)
+            
+            logger.debug(
+                f"Published task {i+1}/{len(request.tasks)} to {agent_name}: "
+                f"task_id={task_id}, batch_id={batch_id}"
+            )
+        
+        logger.info(f"Published {len(task_ids)} tasks to {agent_name} workers with batch_id={batch_id}")
+        
+        if request.wait:
+            # Wait for all responses
+            results = []
+            for task_id in task_ids:
+                response = await message_bus.wait_for_response(task_id, timeout_sec=300)
+                results.append({
+                    "task_id": task_id,
+                    "response": response
+                })
+            
+            return {
+                "batch_id": batch_id,
+                "agent_name": agent_name,
+                "status": "completed",
+                "task_count": len(task_ids),
+                "task_ids": task_ids,
+                "results": results
+            }
+        
+        return {
+            "batch_id": batch_id,
+            "agent_name": agent_name,
+            "status": "submitted",
+            "task_count": len(task_ids),
+            "task_ids": task_ids
+        }
+    
+    except Exception as e:
+        logger.exception("Failed to batch submit tasks")
+        raise HTTPException(status_code=500, detail=str(e))
+
+
 @app.get("/api/agents/{agent_name}/tools")
 async def get_agent_tools(agent_name: str):
     """
@@ -1537,8 +1610,214 @@ async def websocket_prompt_traces(websocket: WebSocket, prompt_id: str):
             pass
 
 
-    @app.post("/api/prompts/{prompt_id}/cancel", response_model=CancelPromptResponse)
-    async def cancel_prompt(prompt_id: str):
+@app.websocket("/ws/batches/{batch_id}")
+async def websocket_batch_traces(websocket: WebSocket, batch_id: str):
+    """
+    WebSocket endpoint for live trace streaming for a batch operation.
+    
+    Continuously fetches traces for all workers in a batch and streams them to the client.
+    Uses batch_id as job_id to group all worker traces together.
+    """
+    try:
+        await websocket.accept()
+        logger.info(f"WebSocket connected for batch {batch_id}")
+        await asyncio.sleep(0.01)
+    except Exception as e:
+        logger.error(f"Failed to accept WebSocket connection for batch {batch_id}: {e}", exc_info=True)
+        return
+    
+    last_trace_id = 0
+    
+    def _parse_ts(ts: str | None) -> datetime | None:
+        if not ts:
+            return None
+        try:
+            return datetime.fromisoformat(ts.replace("Z", "+00:00")).replace(tzinfo=None)
+        except Exception:
+            return None
+    
+    def _build_trace_tree(traces: list[dict]) -> list[dict]:
+        """Build hierarchical trace tree from flat traces."""
+        trace_map = {t['id']: t for t in traces}
+        roots = []
+        
+        # First pass: identify parent-child relationships
+        for trace in traces:
+            parent_id = trace.get('parent_id')
+            if parent_id and parent_id in trace_map:
+                parent = trace_map[parent_id]
+                if 'children' not in parent:
+                    parent['children'] = []
+                parent['children'].append(trace)
+            else:
+                roots.append(trace)
+        
+        # Build span structure
+        def build_span(trace: dict) -> dict:
+            event_type = trace.get('event_type', '')
+            payload = trace.get('payload', {})
+            agent_name = trace.get('agent_name', '')
+            timestamp = trace.get('timestamp', '')
+            
+            # Extract metrics
+            duration_ms = None
+            tokens = None
+            
+            # Calculate duration if we have start/end events
+            if event_type == 'task_complete':
+                start_trace = next((t for t in traces if t.get('agent_name') == agent_name 
+                                   and t.get('event_type') == 'task_start' 
+                                   and t.get('id') < trace['id']), None)
+                if start_trace:
+                    start_ts = _parse_ts(start_trace.get('timestamp'))
+                    end_ts = _parse_ts(timestamp)
+                    if start_ts and end_ts:
+                        duration_ms = int((end_ts - start_ts).total_seconds() * 1000)
+            
+            # Extract token usage
+            if event_type == 'llm_usage' or 'usage' in payload:
+                usage = payload.get('usage', payload)
+                tokens = usage.get('total_tokens', 0)
+            
+            # Build span object
+            span = {
+                'id': trace['id'],
+                'name': agent_name if event_type.startswith('task_') else payload.get('tool', event_type),
+                'type': 'agent' if event_type.startswith('task_') else ('tool' if event_type in ('tool_call', 'tool_result') else ('llm' if event_type in ('llm_call', 'llm_usage') else 'event')),
+                'start_time': timestamp,
+                'agent': agent_name,
+                'event_type': event_type,
+                'input': payload.get('inputs', payload.get('params', {})),
+                'output': payload.get('result', payload.get('outputs')),
+                'metadata': {
+                    'duration_ms': duration_ms,
+                    'tokens': tokens,
+                    **payload
+                },
+                'children': []
+            }
+            
+            # Add children recursively
+            if 'children' in trace:
+                span['children'] = [build_span(child) for child in trace['children']]
+            
+            return span
+        
+        return [build_span(root) for root in roots]
+    
+    # Get event loop once for all executor calls
+    loop = asyncio.get_running_loop()
+    
+    # Helper to safely send WebSocket messages
+    async def safe_send(data: dict) -> bool:
+        try:
+            await websocket.send_json(data)
+            logger.debug(f"Sent WebSocket message type={data.get('type')} for batch {batch_id}")
+            return True
+        except RuntimeError as e:
+            if "closed" in str(e).lower() or "disconnect" in str(e).lower():
+                logger.debug(f"WebSocket connection closed for batch {batch_id}")
+            else:
+                logger.debug(f"RuntimeError sending WebSocket message for batch {batch_id}: {e}")
+            return False
+        except (ConnectionError, OSError) as e:
+            logger.debug(f"WebSocket connection error for batch {batch_id}: {e}")
+            return False
+        except Exception as e:
+            error_msg = str(e).lower()
+            if "close message" in error_msg or "closed" in error_msg:
+                logger.debug(f"WebSocket closed for batch {batch_id}")
+            else:
+                logger.warning(f"Failed to send WebSocket message for batch {batch_id}: {e}")
+            return False
+    
+    # Load initial traces immediately on connection
+    try:
+        initial_traces = await loop.run_in_executor(_db_executor, database.get_job_traces, batch_id)
+        logger.info(f"Loaded {len(initial_traces)} initial traces for batch {batch_id}")
+        if initial_traces:
+            last_trace_id = max(t.get('id', 0) for t in initial_traces)
+            trace_tree = _build_trace_tree(initial_traces)
+            logger.info(f"Built trace tree with {len(trace_tree)} root spans for batch {batch_id}")
+            sent = await safe_send({
+                "type": "traces",
+                "data": {
+                    "spans": trace_tree,
+                    "count": len(initial_traces)
+                }
+            })
+            if sent:
+                logger.info(f"Successfully sent {len(initial_traces)} initial traces for batch {batch_id}")
+            else:
+                logger.debug(f"Could not send initial traces for batch {batch_id} - connection may be closed")
+        else:
+            logger.info(f"No initial traces found for batch {batch_id}")
+            # Send empty traces to let frontend know we're connected
+            await safe_send({
+                "type": "traces",
+                "data": {
+                    "spans": [],
+                    "count": 0
+                }
+            })
+    except Exception as e:
+        logger.error(f"Failed to load initial traces for batch {batch_id}: {e}", exc_info=True)
+    
+    # Poll for new traces continuously
+    try:
+        while True:
+            # Check connection state
+            if websocket.client_state.name != "CONNECTED":
+                logger.info(f"WebSocket disconnected for batch {batch_id}")
+                break
+            
+            # Get all traces for this batch (non-blocking)
+            traces = await loop.run_in_executor(_db_executor, database.get_job_traces, batch_id)
+            
+            # Filter to only new traces
+            new_traces = [t for t in traces if t.get('id', 0) > last_trace_id]
+            
+            if new_traces:
+                # Update last seen ID
+                last_trace_id = max(t.get('id', 0) for t in new_traces)
+                
+                # Build hierarchical tree structure
+                trace_tree = _build_trace_tree(new_traces)
+                
+                logger.debug(f"Sending {len(new_traces)} new traces ({len(trace_tree)} root spans) for batch {batch_id}")
+                
+                # Send tree structure to client
+                sent = await safe_send({
+                    "type": "traces",
+                    "data": {
+                        "spans": trace_tree,
+                        "count": len(new_traces)
+                    }
+                })
+                if sent:
+                    logger.debug(f"Successfully sent {len(new_traces)} new traces for batch {batch_id}")
+                else:
+                    # Connection closed, exit loop
+                    break
+            
+            # Poll interval
+            await asyncio.sleep(0.5)
+    
+    except WebSocketDisconnect:
+        logger.info(f"WebSocket client disconnected for batch {batch_id}")
+    except Exception as e:
+        logger.error(f"WebSocket error for batch {batch_id}: {e}", exc_info=True)
+        try:
+            await websocket.send_json({
+                "type": "error",
+                "data": {"error": str(e)}
+            })
+        except:
+            pass
+
+
+@app.post("/api/prompts/{prompt_id}/cancel", response_model=CancelPromptResponse)
+async def cancel_prompt(prompt_id: str):
         """Request cancellation of a running prompt. This sets a cancel flag and updates status."""
         try:
             # Signal cancel to runtime via message bus
